"""
Calculate the Roll, Pitch and Yaw angles along the trajectory.

Given a selection of three regions defined as:

    selection A or selection B or selection C

calculates the roll, pitch and yaw angles of an axis of reference
calculated around that selection.

The axis of reference is calculated as follows:

    1) the centre of geometry of the selection defines the origin of the
        reference frame
       
       1.1) all frames in the trajectory are centered to that
            origin.

    2) one of the axis of the reference frame is defined by the unitary
        vectir of 'selection A'.

    3) the second axis is defined by the normal vector of the plane
        defined by the centre of geometry of the three selections
        separately.
    
    4) the last axis, is defined by the cross product of the two previous
        axis.

The above procedure is performed for each frame and the reference frame
of the first frame is stored as main reference.

Calculating the angles:

Roll)
    The roll angle is calculated by rotating the 'selection A' unitary
    vector around the 'reference normal vector' until the Quaternion
    distance is the minimum between the 'reference selection A' vector
    and the 'frame i selection A' vector.

Pitch)
    The same procedure as for Roll but the
"""
import argparse

import numpy as np

from taurenmd import log
from taurenmd.libs import libcalc, libcli, libio, libmda, libutil  # noqa: F401
from taurenmd.logger import S, T

_help = 'Calculates angular rotations across axes.'
_name = 'rotations'


ap = libcli.CustomParser(
    description=__doc__,
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    # formatter_class=argparse.RawDescriptionHelpFormatter,
    )

ap.add_argument(
    'topology',
    help='Topology file.',
    type=str,
    )

ap.add_argument(
    'trajectory',
    help=(
        'Trajectory files. If given, multiple trajectories will be'
        'contactenated by order.'
        ),
    nargs='+',
    )

ap.add_argument(
    '--origin-selection',
    help='The selection to define the plane',
    required=True,
    )


ap.add_argument(
    '-s',
    '--start',
    help='Start frame for slicing.',
    default=None,
    type=int,
    )

ap.add_argument(
    '-e',
    '--stop',
    help='Stop frame for slicing: exclusive',
    default=None,
    type=int,
    )

ap.add_argument(
    '-p',
    '--step',
    help='Step value for slicing',
    default=None,
    type=int,
    )


def load_args():
    """Load user arguments."""
    cmd = ap.parse_args()
    return cmd


def maincli():
    cmd = load_args()
    main(**vars(cmd))
    return


def main(
        topology,
        trajectory,
        origin_selection,
        start=None,
        stop=None,
        step=None,
        **kwargs,
        ):
    log.info(T('starting'))
    
    u = libmda.mda_load_universe(topology, *list(trajectory))

    log.info(T('transformation'))
    fSlice = libutil.frame_slice(start=start, stop=stop, step=step)
    
    pABC_atomG = u.select_atoms(origin_selection)
    ABC_selections = [sel.strip() for sel in origin_selection.split('or')]
    # p stands for point
    pA_atomG = u.select_atoms(ABC_selections[0])
    pB_atomG = u.select_atoms(ABC_selections[1])
    pC_atomG = u.select_atoms(ABC_selections[2])

    u.trajectory[0]
    
    # defining the center of reference
    pABC_cog = pABC_atomG.center_of_geometry()
    log.info(T('Original Center of Geometry'))
    log.info(S('for frame: 0'))
    log.info(S('for selection: {}', origin_selection))
    log.info(S('pABC_cog: {}', pABC_cog))
    
    log.info(T('Transfering'))
    log.info(S('all coordinates of reference frame to the origin 0, 0, 0'))
    pABC_atomG.positions = pABC_atomG.positions - pABC_cog
    log.info(S('COG in origin: {}', pABC_atomG.center_of_geometry()))

    log.info(T('defining the reference axes'))
    pA_cog = pA_atomG.center_of_geometry()
    pB_cog = pB_atomG.center_of_geometry()
    pC_cog = pC_atomG.center_of_geometry()
    log.info(S('plane points definition:'))
    log.info(S('pA: {}', pA_cog))
    log.info(S('pB: {}', pB_cog))
    log.info(S('pC: {}', pC_cog))

    log.info(T('defining the normal vector to reference plane'))
    ref_plane_normal = libcalc.calc_plane_normal(pA_cog, pB_cog, pC_cog)
    log.info(S('plane normal: {}', ref_plane_normal))
    log.info(S('done'))

    log.info(T('defining the cross product vector'))
    ref_plane_cross = np.cross(pA_cog, ref_plane_normal)
    log.info(S('np cross: {}', ref_plane_cross))
    
    roll_angles = []
    pitch_angles = []
    yaw_angles = []

    for i, ts in enumerate(u.trajectory[fSlice]):
        print(f'.. working for frame :{i}')

        pABC_cog_ts = pABC_atomG.center_of_geometry()
        pABC_atomG.positions = pABC_atomG.positions - pABC_cog_ts

        pA_cog_ts = pA_atomG.center_of_geometry()
        pB_cog_ts = pB_atomG.center_of_geometry()
        pC_cog_ts = pC_atomG.center_of_geometry()

        ts_plane_normal = libcalc.calc_plane_normal(
            pA_cog_ts,
            pB_cog_ts,
            pC_cog_ts,
            )

        ts_plane_cross = np.cross(pA_cog_ts, ts_plane_normal)

        # Calculating Quaternion Rotations
        roll_Qs_tuples = libcalc.generate_quaternion_rotations(
            ref_plane_normal,
            pA_cog_ts,
            )

        pitch_Qs_tuples = libcalc.generate_quaternion_rotations(
            ref_plane_cross,
            ts_plane_normal,
            )

        yaw_Qs_tuples = libcalc.generate_quaternion_rotations(
            pA_cog,
            ts_plane_cross,
            )

        roll_minimum = libcalc.calc_minimum_Qdistances(roll_Qs_tuples, pA_cog)
        pitch_minimum = \
            libcalc.calc_minimum_Qdistances(pitch_Qs_tuples, ref_plane_normal)
        yaw_minimum = \
            libcalc.calc_minimum_Qdistances(yaw_Qs_tuples, ref_plane_cross)
        
        roll_angles.append(round(roll_minimum.degrees, 3))
        pitch_angles.append(round(pitch_minimum.degrees, 3))
        yaw_angles.append(round(yaw_minimum.degrees, 3))

    print('... saving roll_angles.csv ...')
    libio.save_to_file(
        list(range(1, len(u.trajectory) + 1)),
        [roll_angles],
        fname='roll_angles.csv',
        )

    print('... saving pitch_angles.csv ...')
    libio.save_to_file(
        list(range(1, len(u.trajectory) + 1)),
        [pitch_angles],
        fname='pitch_angles.csv',
        )
    print('... saving yaw_angles.csv ...')
    libio.save_to_file(
        list(range(1, len(u.trajectory) + 1)),
        [yaw_angles],
        fname='yaw_angles.csv',
        )
    log.info(S('done'))
    return


if __name__ == '__main__':
    maincli()



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mtaurenmd` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``taurenmd.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``taurenmd.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import sys

import taurenmd.cli_angle as cli_angle
import taurenmd.cli_distances as cli_dist
import taurenmd.cli_fext as cli_fext
import taurenmd.cli_imagemol as cli_imagemol
import taurenmd.cli_noSol as cli_noSol
import taurenmd.cli_report as cli_report
import taurenmd.cli_rmsd as cli_rmsd
import taurenmd.cli_rmsf as cli_rmsf
import taurenmd.cli_rotations as cli_rot
import taurenmd.cli_trajedit as cli_trajedit
from taurenmd import CMDFILE, log
from taurenmd.libs import libcli


ap = libcli.CustomParser(description=__doc__)
subparsers = ap.add_subparsers(title='TAURENMD SUBROUTINES')

libcli.add_subparser(subparsers, cli_angle)
libcli.add_subparser(subparsers, cli_dist)
libcli.add_subparser(subparsers, cli_fext)
libcli.add_subparser(subparsers, cli_imagemol)
libcli.add_subparser(subparsers, cli_noSol)
libcli.add_subparser(subparsers, cli_report)
libcli.add_subparser(subparsers, cli_rmsd)
libcli.add_subparser(subparsers, cli_rmsf)
libcli.add_subparser(subparsers, cli_rot)
libcli.add_subparser(subparsers, cli_trajedit)


#ap_angle = subparsers.add_parser(
#    'angle',
#    description=cli_angle.__doc__,
#    help='Calculates the angle between a plane along the trajectory.',
#    parents=[cli_angle.ap],
#    add_help=False,
#    )
#ap_angle.set_defaults(func=cli_angle.main)

#ap_dist = subparsers.add_parser(
#    'dist',
#    description=cli_dist.__doc__,
#    help='Calculates distances between geometric centres of selections',
#    parents=[cli_dist.ap],
#    add_help=False,
#    )
#ap_dist.set_defaults(func=cli_dist.main)
#
#ap_imagemol = subparsers.add_parser(
#    'imagemol',
#    description=cli_imagemol.__doc__,
#    help='Attempts to image molecules.',
#    parents=[cli_imagemol.ap],
#    add_help=False,
#    )
#ap_imagemol.set_defaults(func=cli_imagemol.main)
#
#ap_fext = subparsers.add_parser(
#    'fext',
#    description=cli_fext.__doc__,
#    help='Extracts single frames from trajectory.',
#    parents=[cli_fext.ap],
#    add_help=False,
#    )
#ap_fext.set_defaults(func=cli_fext.main)
#
#ap_noSol = subparsers.add_parser(
#    'noSol',
#    description=cli_noSol.__doc__,
#    help='Removes solvent and extracts fist frame',
#    parents=[cli_noSol.ap],
#    add_help=False,
#    )
#ap_noSol.set_defaults(func=cli_noSol.main)
#
#ap_tedit = subparsers.add_parser(
#    'trajedit',
#    description=cli_trajedit.__doc__,
#    help='Converts traj to PDB.',
#    parents=[cli_trajedit.ap],
#    add_help=False,
#    )
#ap_tedit.set_defaults(func=cli_trajedit.main)
#
#ap_report = subparsers.add_parser(
#    'report',
#    description=cli_report.__doc__,
#    help='Reports on traj details.',
#    parents=[cli_report.ap],
#    add_help=False,
#    )
#ap_report.set_defaults(func=cli_report.main)
#
#ap_rot = subparsers.add_parser(
#    'rotations',
#    description=cli_rot.__doc__,
#    help='Calculates angular rotations across axes.',
#    parents=[cli_rot.ap],
#    add_help=False,
#    )
#ap_rot.set_defaults(func=cli_rot.main)
#
#ap_rmsd = subparsers.add_parser(
#    'rmsd',
#    description=cli_rmsd.__doc__,
#    help='Calculates and plot RMSDs',
#    parents=[cli_rmsd.ap],
#    add_help=False,
#    )
#ap_rmsd.set_defaults(func=cli_rmsd.main)
#
#ap_rmsf = subparsers.add_parser(
#    'rmsf',
#    description=cli_rmsf.__doc__,
#    help='Calculates and plot RMSFs',
#    parents=[cli_rmsf.ap],
#    add_help=False,
#    )
#ap_rmsf.set_defaults(func=cli_rmsf.main)


def _ap():
    return ap


def load_args():
    return ap.parse_args()


def maincli():
    libcli.save_command(CMDFILE, *sys.argv)

    if len(sys.argv) < 2:
        ap.print_help()
        ap.exit()

    args = load_args()
    log.debug(args)
    args.func(**vars(args))


if __name__ == '__main__':
    maincli()
